@@ -48,6 +48,9 @@
 #define PARENT_TYPE MAIL_TYPE_STUB
 static MailStubClass *parent_class = NULL;
 
+/* FIXME : Have this as part of the appropriate class in 2.5 */
+ExchangeOfflineListener *listener = NULL;
+
 typedef struct {
 	char *uid, *href;
 	guint32 seq, flags;
@@ -129,7 +132,7 @@ static void linestatus_listener (ExchangeOfflineListener *listener,
 						    gint linestatus,
 						    gpointer data);
 static void folder_update_linestatus (gpointer key, gpointer value, gpointer data);
-
+static void free_folder (gpointer value);
 
 static void
 class_init (GObjectClass *object_class)
@@ -168,7 +171,8 @@ init (GObject *object)
 {
 	MailStubExchange *mse = MAIL_STUB_EXCHANGE (object);
 
-	mse->folders_by_name = g_hash_table_new (g_str_hash, g_str_equal);
+	mse->folders_by_name = g_hash_table_new_full (g_str_hash, g_str_equal, 
+						      NULL, free_folder);
 }
 
 static void
@@ -181,7 +185,7 @@ free_message (MailStubExchangeMessage *mmsg)
 }
 
 static void
-free_folder (gpointer key, gpointer value, gpointer ctx)
+free_folder (gpointer value)
 {
 	MailStubExchangeFolder *mfld = value;
 	int i;
@@ -212,8 +216,6 @@ dispose (GObject *object)
 	MailStubExchange *mse = MAIL_STUB_EXCHANGE (object);
 
 	if (mse->folders_by_name) {
-		g_hash_table_foreach (mse->folders_by_name, free_folder,
-				      mse->ctx);
 		g_hash_table_destroy (mse->folders_by_name);
 		mse->folders_by_name = NULL;
 	}
@@ -230,6 +232,11 @@ dispose (GObject *object)
 		mse->removed_folder_id = 0;
 	}
 
+	if (listener) {
+		g_signal_handlers_disconnect_by_func (listener, linestatus_listener, mse);
+		listener = NULL;
+	}
+
 	G_OBJECT_CLASS (parent_class)->dispose (object);
 }
 
@@ -514,13 +521,14 @@ static void
 got_folder_error (MailStubExchangeFolder *mfld, const char *error)
 {
 	mail_stub_return_error (MAIL_STUB (mfld->mse), error);
-	free_folder (NULL, mfld, NULL);
+	free_folder (mfld);
 }
 
 static const char *open_folder_sync_props[] = {
 	E2K_PR_REPL_UID,
 	PR_INTERNET_ARTICLE_NUMBER,
 	PR_ACTION_FLAG,
+	PR_IMPORTANCE,
 	PR_DELEGATED_BY_RULE,
 	E2K_PR_HTTPMAIL_READ,
 	E2K_PR_HTTPMAIL_MESSAGE_FLAG,
@@ -987,6 +995,7 @@ static const char *new_message_props[] = {
 	E2K_PR_HTTPMAIL_READ,
 	E2K_PR_HTTPMAIL_HAS_ATTACHMENT,
 	PR_ACTION_FLAG,
+	PR_IMPORTANCE,
 	PR_DELEGATED_BY_RULE,
 	E2K_PR_HTTPMAIL_MESSAGE_FLAG,
 	E2K_PR_MAILHEADER_REPLY_BY,
@@ -1460,6 +1469,28 @@ set_replied_flags (MailStubExchange *mse, MailStubExchangeMessage *mmsg)
 		g_warning ("set_replied_flags: %d", status);
 }
 
+static void
+set_important_flag (MailStubExchange *mse, MailStubExchangeMessage *mmsg)
+{
+	E2kProperties *props;
+	E2kHTTPStatus status;
+
+	props = e2k_properties_new ();
+
+	if (mmsg->change_flags & MAIL_STUB_MESSAGE_FLAGGED) {
+		e2k_properties_set_int (props, PR_IMPORTANCE, MAPI_IMPORTANCE_HIGH);
+	}
+	else {
+		e2k_properties_set_int (props, PR_IMPORTANCE, MAPI_IMPORTANCE_NORMAL);
+	}
+
+	status = e2k_context_proppatch (mse->ctx, NULL, mmsg->href, props,
+					FALSE, NULL);
+	e2k_properties_free (props);
+	if (!E2K_HTTP_STATUS_IS_SUCCESSFUL (status))
+		g_warning ("set_important_flag: %d", status);
+}
+
 static void
 update_tags (MailStubExchange *mse, MailStubExchangeMessage *mmsg)
 {
@@ -1551,6 +1582,11 @@ process_flags (gpointer user_data)
 			mmsg->change_mask &= ~(MAIL_STUB_MESSAGE_ANSWERED | MAIL_STUB_MESSAGE_ANSWERED_ALL);
 		}
 
+		if (mmsg->change_mask & MAIL_STUB_MESSAGE_FLAGGED) {
+			set_important_flag (mse, mmsg);
+			mmsg->change_mask &= ~MAIL_STUB_MESSAGE_FLAGGED;
+		}
+
 		if (mmsg->tag_updates)
 			update_tags (mse, mmsg);
 
@@ -2148,10 +2184,10 @@ get_folder_info (MailStub *stub, const char *top, guint32 store_flags)
 	GArray *unread, *flags;	
 	ExchangeHierarchy *hier;
 	EFolder *folder;
-	const char *type, *name, *uri, *path;
+	const char *type, *name, *uri, *path, *inbox_uri;
 	int unread_count, i, toplen = top ? strlen (top) : 0;
 	guint32 folder_flags;
-	gboolean recursive, subscribed;
+	gboolean recursive, subscribed, single_folder = FALSE;
 
 	recursive = (store_flags & CAMEL_STUB_STORE_FOLDER_INFO_RECURSIVE);
 	subscribed = (store_flags & CAMEL_STUB_STORE_FOLDER_INFO_SUBSCRIBED);
@@ -2162,15 +2198,20 @@ get_folder_info (MailStub *stub, const char *top, guint32 store_flags)
 		exchange_account_open_folder (mse->account, "/public");
 	}
 
-	if (!recursive && subscribed) {
+	if (!recursive && toplen) {
 		char *full_path;
 
 		full_path = g_strdup_printf ("/%s", top);
 		folder = exchange_account_get_folder (mse->account, full_path);
 		g_free (full_path);
 		if (folder) {
-			folders = g_ptr_array_new ();
-			g_ptr_array_add (folders, folder);
+			type = e_folder_get_type_string (folder);
+			if ((!strcmp (type, "mail")) || 
+				(!strcmp (type, "mail/public"))) {
+				folders = g_ptr_array_new ();
+				g_ptr_array_add (folders, folder);
+				single_folder = TRUE;
+			}
 		}
 	} else {
 		folders = exchange_account_get_folders (mse->account);
@@ -2180,7 +2221,7 @@ get_folder_info (MailStub *stub, const char *top, guint32 store_flags)
 	uris = g_ptr_array_new ();
 	unread = g_array_new (FALSE, FALSE, sizeof (int));
 	flags = g_array_new (FALSE, FALSE, sizeof (int));
-	//flags = g_byte_array_new ();
+	inbox_uri = e_folder_get_physical_uri (mse->inbox);
 
 	if (folders) {
 		for (i = 0; i < folders->len; i++) {
@@ -2188,6 +2229,10 @@ get_folder_info (MailStub *stub, const char *top, guint32 store_flags)
 			hier = e_folder_exchange_get_hierarchy (folder);
 			folder_flags = 0;
 
+			/* We ll have only that folder in the folders array */
+			if (single_folder)
+				goto return_data;
+
 			if (subscribed) {
 				if (hier->type != EXCHANGE_HIERARCHY_PERSONAL &&
 				    hier->type != EXCHANGE_HIERARCHY_FAVORITES &&
@@ -2198,33 +2243,52 @@ get_folder_info (MailStub *stub, const char *top, guint32 store_flags)
 					continue;
 			}
 
-			if (recursive && top) {
+			if (recursive && toplen) {
 				path = e_folder_exchange_get_path (folder);
 				if (strncmp (path + 1, top, toplen) != 0)
 					continue;
 			}
-
+return_data:
 			type = e_folder_get_type_string (folder);
 			name = e_folder_get_name (folder);
 			uri = e_folder_get_physical_uri (folder);
-			d(printf ("folder type is : %s\n", type));
-			if ((!strcmp (type, "mail")) || 
-				(!strcmp (type, "mail/public")) ||
-				(!strcmp (type, "calendar/public")) ||
-				(!strcmp (type, "tasks/public")) ||
-				(!strcmp (type, "contacts/public"))) {
-				unread_count = e_folder_get_unread_count (folder);
-				if (exchange_account_is_favorite_folder (mse->account, folder)) {
-					folder_flags |= CAMEL_STUB_FOLDER_SUBSCRIBED;
-					d(printf ("marked the folder as subscribed\n"));
-				}
-				else
-					folder_flags = 0;
-			} else {
+			d(g_print ("Uri: %s\n", uri));
+			d(g_print ("folder type is : %s\n", type));
+
+			if (!strcmp (type, "noselect")) {
 				unread_count = 0;
 				folder_flags = CAMEL_STUB_FOLDER_NOSELECT;
 			}
-			if (folder == mse->inbox)
+
+			switch (hier->type) {
+				case EXCHANGE_HIERARCHY_FAVORITES:
+					/* folder_flags will be set only if the type
+					   is noselect and we need to include it */
+					if (strcmp (type, "mail") && !folder_flags)
+						continue;
+					/* selectable */
+					if (!folder_flags)
+						unread_count = e_folder_get_unread_count (folder);
+				case EXCHANGE_HIERARCHY_PUBLIC:
+					if (exchange_account_is_favorite_folder (mse->account, folder)) {
+						folder_flags |= CAMEL_STUB_FOLDER_SUBSCRIBED;
+						d(printf ("marked the folder as subscribed\n"));
+					}
+					break;
+				case EXCHANGE_HIERARCHY_FOREIGN:
+				case EXCHANGE_HIERARCHY_PERSONAL:
+					if (!strcmp (type, "mail")) {
+						unread_count = e_folder_get_unread_count (folder);
+					}
+					else if(!folder_flags) {
+						continue;
+					}
+					break;
+				default:
+					break;
+			}
+
+			if (!strcmp (uri, inbox_uri))
 				folder_flags |= CAMEL_STUB_FOLDER_SYSTEM|CAMEL_STUB_FOLDER_TYPE_INBOX;
 
 			g_ptr_array_add (names, (char *)name);
@@ -2392,6 +2456,7 @@ delete_folder (MailStub *stub, const char *folder_name)
 	switch (result) {
 	case EXCHANGE_ACCOUNT_FOLDER_OK:
 	case EXCHANGE_ACCOUNT_FOLDER_DOES_NOT_EXIST:
+		g_hash_table_remove (mse->folders_by_name, folder_name);
 		break;
 
 	case EXCHANGE_ACCOUNT_FOLDER_PERMISSION_DENIED:
@@ -2446,8 +2511,14 @@ rename_folder (MailStub *stub, const char *old_name, const char *new_name)
 		mfld->folder = g_object_ref (folder);
 		mfld->name = e_folder_exchange_get_path (folder) + 1;
 
-		g_hash_table_remove (mse->folders_by_name, old_name);
+		g_hash_table_steal (mse->folders_by_name, old_name);
 		g_hash_table_insert (mse->folders_by_name, (char *)mfld->name, mfld);
+
+		mail_stub_return_data (stub, CAMEL_STUB_RETVAL_FOLDER_RENAMED,
+				       CAMEL_STUB_ARG_FOLDER, old_name,
+				       CAMEL_STUB_ARG_FOLDER, e_folder_get_name (folder),
+				       CAMEL_STUB_ARG_STRING, e_folder_get_physical_uri (folder),
+				       CAMEL_STUB_ARG_END);
 		break;
 
 	case EXCHANGE_ACCOUNT_FOLDER_DOES_NOT_EXIST:
@@ -2482,10 +2553,16 @@ subscribe_folder (MailStub *stub, const char *folder_name)
 		g_free (path);
 		return;
 	}
+	g_free (path);
 	g_object_ref (folder);
 
+	if (e_folder_exchange_get_hierarchy (folder)->type != EXCHANGE_HIERARCHY_PUBLIC) {
+		g_object_unref (folder);
+		mail_stub_return_ok (stub);
+		return;
+	}
+	
 	result = exchange_account_add_favorite (mse->account, folder);
-	g_free (path);
 
 	switch (result) {
 	case EXCHANGE_ACCOUNT_FOLDER_OK:
@@ -2501,12 +2578,10 @@ subscribe_folder (MailStub *stub, const char *folder_name)
 		mail_stub_return_error (stub, _("Generic error"));
 		g_object_unref (folder);
 		return;
-
 	}
 
 	g_object_unref (folder);
 	mail_stub_return_ok (stub);
-
 }
 
 static void
@@ -2516,8 +2591,7 @@ unsubscribe_folder (MailStub *stub, const char *folder_name)
 	MailStubExchange *mse = MAIL_STUB_EXCHANGE (stub);
 	ExchangeAccountFolderResult result;
 	EFolder *folder;
-	char *path;
-	const char *folder_type, *physical_uri;
+	char *path, *pub_name;
 
 	path = g_build_filename ("/", folder_name, NULL);
 	folder = exchange_account_get_folder (mse->account, path);
@@ -2526,14 +2600,37 @@ unsubscribe_folder (MailStub *stub, const char *folder_name)
 		g_free (path);
 		return;
 	}
+	g_free (path);
+	g_object_ref (folder);
+
+	/* if (e_folder_exchange_get_hierarchy (folder)->type != EXCHANGE_HIERARCHY_FAVORITES) {
+	   Should use above check, but the internal uri is the same for both 
+	   public and favorite hierarchies and any of them can be used for the check */
+	if (!exchange_account_is_favorite_folder (mse->account, folder)) {
+		g_object_unref (folder);
+		mail_stub_return_ok (stub);
+		return;
+	}
+
+	g_object_unref (folder);
+	
+	pub_name = strchr (folder_name, '/');
+	path = g_build_filename ("/favorites", pub_name, NULL);
+	folder = exchange_account_get_folder (mse->account, path);
+	if (!folder) {
+		mail_stub_return_error (stub, _("Folder doesn't exist"));
+		g_free (path);
+		return;
+	}
+	g_free (path);
 	g_object_ref (folder);
 
 	result = exchange_account_remove_favorite (mse->account, folder);
-	g_free (path);
 
 	switch (result) {
 	case EXCHANGE_ACCOUNT_FOLDER_OK:
 	case EXCHANGE_ACCOUNT_FOLDER_DOES_NOT_EXIST:
+		g_hash_table_remove (mse->folders_by_name, path + 1);
 		break;
 
 	case EXCHANGE_ACCOUNT_FOLDER_PERMISSION_DENIED:
@@ -2548,40 +2645,6 @@ unsubscribe_folder (MailStub *stub, const char *folder_name)
 
 	}
 
-	if (result == EXCHANGE_ACCOUNT_FOLDER_OK) {
-		/* This is a dirty hack to remove the esources for favourite
-			folders. ExchangeConfigListener depends on EAccountList
-			which is present only in evolution, and so cannot be
-			included as part of the servers code in e-d-s. So,
-			ExchangeConfigListener had to be moved to evolution 
-			plugins and evolution-exchange. Ideally, this esource
-			handling should have been done in the favorites code.
-			The following code has been moved from 
-			exchanga-hierarchy-favorites.c:remove_folder
-		*/
-		/* Temp Fix for remove fav folders. see #59168 */
-		/* remove ESources */
-		folder_type = e_folder_get_type_string (folder);
-		physical_uri = e_folder_get_physical_uri (folder);
-		d(printf ("removeing folder type : %s\n", folder_type));
-		d(printf ("physical uri : %s \n", physical_uri));
-
-		if (strcmp (folder_type, "calendar/public") == 0) {
-			remove_folder_esource (mse->account,
-					       EXCHANGE_CALENDAR_FOLDER,
-					       physical_uri);
-		}
-		else if (strcmp (folder_type, "tasks/public") == 0) {
-			remove_folder_esource (mse->account,
-					       EXCHANGE_TASKS_FOLDER,
-					       physical_uri);
-		}
-		else if (strcmp (folder_type, "contacts/public") == 0) {
-			remove_folder_esource (mse->account,
-					       EXCHANGE_CONTACTS_FOLDER,
-					       physical_uri);
-		}
-	}
 	g_object_unref (folder);
 	mail_stub_return_ok (stub);
 }
@@ -2590,7 +2653,6 @@ static void
 is_subscribed_folder (MailStub *stub, const char *folder_name)
 {
 	MailStubExchange *mse = MAIL_STUB_EXCHANGE (stub);
-	gboolean result;
 	EFolder *folder;
 	char *path;
 	guint32 is_subscribed = 0;
@@ -2598,20 +2660,19 @@ is_subscribed_folder (MailStub *stub, const char *folder_name)
 	path = g_build_filename ("/", folder_name, NULL);
 	folder = exchange_account_get_folder (mse->account, path);
 	if (!folder) {
-		/* Dont ever return an exception from here as CamelException
-		is NULL for this */
-		//mail_stub_return_error (stub, _("Folder doesn't exist"));
 		g_free (path);
+		mail_stub_return_data (stub, CAMEL_STUB_RETVAL_RESPONSE,
+				       CAMEL_STUB_ARG_UINT32, is_subscribed,
+				       CAMEL_STUB_ARG_END);
+		mail_stub_return_ok (stub);
 		return;
 	}
+	g_free (path);
 	g_object_ref (folder);
 
-	result = exchange_account_is_favorite_folder (mse->account, folder);
-	g_free (path);
-
-	if (result)
+	if (exchange_account_is_favorite_folder (mse->account, folder))
 		is_subscribed = 1;
-	
+
 	g_object_unref (folder);
 
 	mail_stub_return_data (stub, CAMEL_STUB_RETVAL_RESPONSE,
@@ -2635,7 +2696,7 @@ linestatus_listener (ExchangeOfflineListener *listener,
 	ExchangeAccount *account = mse->account;
 	const char *uri;
 	
-	if (linestatus == ONLINE_MODE) {
+	if (listener && (linestatus == ONLINE_MODE)) {
 		mse->ctx = exchange_account_get_context (account);
 		g_object_ref (mse->ctx);
 
@@ -2716,7 +2777,6 @@ mail_stub_exchange_new (ExchangeAccount *account, int cmd_fd, int status_fd)
 	MailStub *stub;
 	const char *uri;
 	int mode;
-	ExchangeOfflineListener *listener;
 
 	stub = g_object_new (MAIL_TYPE_STUB_EXCHANGE, NULL);
 	g_object_ref (stub);
@@ -2738,7 +2798,7 @@ mail_stub_exchange_new (ExchangeAccount *account, int cmd_fd, int status_fd)
 	
 	listener = exchange_component_get_offline_listener (global_exchange_component);
 	g_signal_connect (G_OBJECT (listener), "linestatus-changed",
-			  G_CALLBACK (linestatus_listener), mse);
+				 G_CALLBACK (linestatus_listener), mse);
 
 	return stub;
 }
